package soot.javaToJimple.ppa;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class PPAReport {

	private static PPAReport instance = new PPAReport();
	
	private Map<String,Integer> firstStrategy = new HashMap<String,Integer>();
	
	private Map<String,Integer> totalStrategy = new HashMap<String,Integer>();
	
	private Map<Integer,Integer> inferenceLengths = new HashMap<Integer,Integer>();
	
	private long factCount = 0;
	
	private long total = 0;
	
	private long lengthCount = 0;
	
	private Set<TypeFact> inferedFacts = new HashSet<TypeFact>();
	
	
	
	public static PPAReport v() {
		return instance;
	}
	
	public void reportTypeFact(TypeFact fact) {
		factCount++;
		
		addFact(fact,firstStrategy);
		addFact(fact,totalStrategy);
		inferedFacts.add(fact);
		
		TypeFact temp = fact;
		while (temp.getInferenceChain().size() > 0) {
			temp = temp.getInferenceChain().get(0);
			addFact(temp,totalStrategy);
			
			if (inferedFacts.contains(temp)) {
				inferedFacts.remove(temp);
			}
		}
		
	}
	
	private void addFact(TypeFact fact, Map<String, Integer> map) {
		String strategy = fact.getStrategy();
		Integer count = map.get(strategy);
		if (count == null) {
			map.put(strategy, 1);
		} else {
			map.put(strategy, count+1);
		}
	}

	public void printReport() {
		System.out.println("REPORT!");
		System.out.println();
		System.out.println("Total facts: " + factCount);
		System.out.println();
		System.out.println("Strategies");
		for (String key: totalStrategy.keySet()) {
			int first = 0;
			if (firstStrategy.containsKey(key)) {
				first = firstStrategy.get(key);
			}
			int total = totalStrategy.get(key);
			System.out.println("  " + key + " : " + first + " (" + total + ")");
		}
		System.out.println();
		
		computeLengths();
		
		System.out.println("Average: " + ((double)total/(double)lengthCount));
		
		for (Integer value : inferenceLengths.keySet()) {
			System.out.println(value + " : " + inferenceLengths.get(value));
		}
		
		
	}
	
	private void computeLengths() {
		for (TypeFact fact : inferedFacts) {
			lengthCount++;
			int length = length(fact);
			Integer count = inferenceLengths.get(length);
			if (count == null) {
				inferenceLengths.put(length, 1);
			} else {
				inferenceLengths.put(length, count + 1);
			}
			total+=length;
		}
	}
	
	private int length(TypeFact typeFact) {
		int length = 1;
		TypeFact temp = typeFact;
		while (temp.getInferenceChain().size() > 0) {
			length++;
			temp = temp.getInferenceChain().get(0);
		}
		return length;
	}
	
	public void reset() {
		firstStrategy.clear();
		inferedFacts.clear();
		inferenceLengths.clear();
		totalStrategy.clear();
		factCount = 0;
		total = 0;
	}
	
	public void softReset() {
		computeLengths();
		inferedFacts.clear();
	}
	
	
	
	
}
