package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.jj.ast.JjArrayInit_c;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;

public class PPAArrayInit_c extends JjArrayInit_c implements PPANode {

	public PPAArrayInit_c(Position pos, List elements) {
		super(pos, elements);
	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;

		try {
			node = super.typeCheck(tc);
		} catch (SemanticException se) {
			Type type = inferArgsType();
			if (type != null) {
				type = tc.typeSystem().arrayOf(type);
			} else {
				type = tc.typeSystem().Null();
			}

			if (!isSafe() && !type.isNull()) {
				reportSaferArgs(type.toArray().base(), null);
			}
			node = type(type);
		}

		return node;
	}

	private void reportSaferArgs(Type type, TypeFact origin) {
		for (Iterator i = elements.iterator(); i.hasNext();) {
			Expr node = (Expr) i.next();
			Type eType = node.type();
			if (TypeFactUtil.isTypeSafer(type, eType)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(node), eType, type,
								TypeFact.CHILD_OR_EQUALS, TypeFact.ARRAY_STRATEGY, origin));
			}
		}

	}

	private Type inferArgsType() {
		Type type = null;
		Type toCompare = null;
		for (Iterator i = elements.iterator(); i.hasNext();) {
			Expr node = (Expr) i.next();
			toCompare = node.type();
			if (type == null) {
				type = toCompare;
			} else if (TypeFactUtil.isTypeSafer(toCompare, type)) {
				type = toCompare;
			}
		}

		return type;
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		
		indexes.add(getMainIndex());
		
		for (Iterator i = elements.iterator(); i.hasNext();) {
			Object node = i.next();
			PPAIndex index = TypeFactUtil.getMainIndex(node);
			if (index != null) {
				indexes.add(index);
			}
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		boolean isSafe = true;
		for (Iterator i = elements.iterator(); i.hasNext();) {
			Object node = i.next();
			isSafe = TypeFactUtil.isSafe(node);
			if (!isSafe) {
				break;
			}
		}
		return isSafe;
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		if (index.equals(getMainIndex())) {
			Type newType = origin.getNewType();
			if (TypeFactUtil.isTypeSafer(newType, type) && !newType.isNull() && newType.isArray()) {
				reportSaferArgs(newType.toArray().base(), origin);
				type = newType;
			}
		} else {
			Type newType = inferArgsType();
			if (newType != null) {
				newType = PPAEngine.v().getTypeSystem().arrayOf(newType);
			}
			
			if (newType != null && TypeFactUtil.isTypeSafer(newType,type)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(getMainIndex(), type, newType,
								TypeFact.CHILD_OR_EQUALS, TypeFact.ARRAY_STRATEGY, origin));
				type = newType;
			}

			if (!isSafe() && !type.isNull() && type.isArray()) {
				reportSaferArgs(type.toArray().base(), null);
			}
		}

	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPAArrayInit_c) {
			PPAArrayInit_c array = (PPAArrayInit_c) obj;
			return TypeFactUtil.positionEquals(position, array.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}

}
