package soot.javaToJimple.ppa.jj.ast;

import java.util.Iterator;
import java.util.List;

import polyglot.ast.ClassBody;
import polyglot.ast.Node;
import polyglot.ast.TypeNode;
import polyglot.ext.jl.ast.ClassDecl_c;
import polyglot.main.Report;
import polyglot.types.Context;
import polyglot.types.Flags;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.visit.AmbiguityRemover;
import soot.javaToJimple.ppa.jj.types.MagicClass;

public class PPAClassDecl_c extends ClassDecl_c {
	public PPAClassDecl_c(Position pos, Flags flags, String name, TypeNode superClass,
			List interfaces, ClassBody body) {
		super(pos, flags, name, superClass, interfaces, body);
	}

	@Override
	public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
		if (ar.kind() == AmbiguityRemover.SIGNATURES) {
			// make sure that the inStaticContext flag of the class is
			// correct
			Context ctxt = ar.context();
			this.type().inStaticContext(ctxt.inStaticContext());

		}

		if (ar.kind() == AmbiguityRemover.SIGNATURES) {
			// make sure that this class has the correct dependencies
			// recorded for its super classes and interfaces.
			ar.addSuperDependencies(this.type());
		}

		if (ar.kind() != AmbiguityRemover.SUPER) {
			return this;
		}

		TypeSystem ts = ar.typeSystem();

		if (Report.should_report(Report.types, 2))
			Report.report(2, "Cleaning " + type + ".");

		disambiguateSuperType(ar);

		for (Iterator i = this.interfaces.iterator(); i.hasNext();) {
			TypeNode tn = (TypeNode) i.next();
			Type t = tn.type();

			if (!t.isCanonical()) {
				throw new SemanticException("Could not disambiguate super" + " class of " + type
						+ ".", tn.position());
			}

			if (!t.isClass() || !t.toClass().flags().isInterface()) {
				if (t instanceof MagicClass) {
					MagicClass mClass = (MagicClass) t;
					mClass.setFlags(mClass.flags().Interface());
				} else {
					throw new SemanticException("Interface " + t + " of " + type
							+ " is not an interface.", tn.position());
				}
			}

			if (Report.should_report(Report.types, 3))
				Report.report(3, "adding interface of " + this.type + " to " + t);

			if (!this.type.interfaces().contains(t))
				this.type.addInterface(t);

			ts.checkCycles(t.toReference());
		}

		return this;
	}
}
