package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Node;
import polyglot.ast.Receiver;
import polyglot.ast.Special;
import polyglot.ext.jl.ast.CanonicalTypeNode_c;
import polyglot.ext.jl.ast.Field_c;
import polyglot.ext.jl.types.FieldInstance_c;
import polyglot.types.Context;
import polyglot.types.FieldInstance;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.InternalCompilerError;
import polyglot.util.Position;
import polyglot.visit.NodeVisitor;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;
import soot.javaToJimple.ppa.jj.types.MagicClass;

/**
 * <p>
 * Field
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li></li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>field instance</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Field instance type != MMAGICPPACKAGE.MagicClass</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPAField_c extends Field_c implements PPANode {

	@Override
	public Node visitChildren(NodeVisitor v) {
		if (v instanceof TypeChecker) {
			TypeChecker tc = (TypeChecker) v;
			// TODO CHECK THIS HACK!
			// if (name.equals("PyType")) {
			// System.out.println();
			// }
			//			

			// This is when you access a field from a static context: the field should be static and
			// is probably not declared in this class.
//			if (tc.context().inStaticContext() && target instanceof Special) {
//				target = new CanonicalTypeNode_c(position, fi.container());
//			}
		}
		return super.visitChildren(v);
	}

	public PPAField_c(Position pos, Receiver target, String name) {
		super(pos, target, name);
	}
	
	public void setTarget(Receiver target) {
		this.target = target;
	}

	public List<PPAIndex> getIndexes() {

		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		indexes.add(new PPAIndex(fieldInstance()));

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(fieldInstance());
	}

	public boolean isSafe() {
		Type fieldType = type();
		// TODO CHECK IF THIS CONDITION MAKES SENSE (PROBABLY...)
		// Yes, it makes sense in the sense that once it is assigned, it cannot change.
		// This is a condition for PPAAssignHelper to work.
		// This could be changed if PPAAssignHelper gave priority to a new type fact.
		return !TypeFactUtil.isAnonymousMagicClass(fieldType)
				&& !TypeFactUtil.isAnonymousArray(fieldType);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		FieldInstance oldFi = fieldInstance();

		FieldInstance newFi = new FieldInstance_c(oldFi.typeSystem(), oldFi.position(), oldFi
				.container(), oldFi.flags(), newType, oldFi.name());

		// Only change the field instance parent if it is a magic class.
		// This should always be the case.
		// TODO (Less Important) Check if this condition is necessary.
		if (oldFi.container() instanceof MagicClass) {
			MagicClass container = (MagicClass) oldFi.container();
			container.replaceField(oldFi, newFi);
			this.fi = newFi;
		}

		type = newType;

		// return true;
	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		// TODO Auto-generated method stub
		return super.typeCheck(tc);
	}

	@Override
	protected void checkConsistency(Context c) {
		try {
			super.checkConsistency(c);
		} catch (InternalCompilerError se) {
			// Will need to do something here, but not for now!
			System.err.println("There was a problem in PPAField.");
			se.printStackTrace();
		}
	}

}
