package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Local;
import polyglot.ast.Node;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.jj.ast.JjLocalAssign_c;
import soot.javaToJimple.ppa.PPAAssignHelper;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;

/**
 * <p>Local variable assignment. E.g., local = 2;</p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li>Based on PPAAssignHelper</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>the node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>left node index</li>
 * <li>right node index</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Left and Right nodes are safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 *
 */
public class PPALocalAssign_c extends JjLocalAssign_c implements PPANode {
	
	public PPALocalAssign_c(Position pos, Local left, Operator op, Expr right) {
		super(pos, left, op, right);
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();

		indexes.add(getMainIndex());

		if (!TypeFactUtil.isSafe(left)) {
			indexes.add(TypeFactUtil.getMainIndex(left));
		}

		if (!TypeFactUtil.isSafe(right)) {
			indexes.add(TypeFactUtil.getMainIndex(right));
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(left) && TypeFactUtil.isSafe(right);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		try {
			PPAAssignHelper.inferTypeInfo(left, right, op, null, newType, origin);
			type = newType;
		} catch (Exception e) {
		}

	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;
		SemanticException se = null;
		try {
			node = super.typeCheck(tc);
		} catch (SemanticException seTemp) {
			se = seTemp;
		}

		PPAAssignHelper.inferTypeInfo(left, right, op, se, null, null);

		if (node == null) {
			node = type(left.type());
		}

		return node;
	}
	
	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPALocalAssign_c) {
			PPALocalAssign_c ppaObj = (PPALocalAssign_c) obj;
			return TypeFactUtil.positionEquals(position, ppaObj.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}
}
