/* Soot - a J*va Optimization Framework
 * Copyright (C) 2004 Jennifer Lhotak
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package soot.javaToJimple.ppa.jj.ast;

import java.util.List;

import polyglot.ast.AmbTypeNode;
import polyglot.ast.ArrayAccess;
import polyglot.ast.ArrayAccessAssign;
import polyglot.ast.ArrayInit;
import polyglot.ast.Assign;
import polyglot.ast.Binary;
import polyglot.ast.Block;
import polyglot.ast.Call;
import polyglot.ast.Case;
import polyglot.ast.Cast;
import polyglot.ast.ClassBody;
import polyglot.ast.ClassDecl;
import polyglot.ast.Conditional;
import polyglot.ast.ConstructorCall;
import polyglot.ast.ConstructorDecl;
import polyglot.ast.Disamb;
import polyglot.ast.Do;
import polyglot.ast.Expr;
import polyglot.ast.Field;
import polyglot.ast.FieldAssign;
import polyglot.ast.FieldDecl;
import polyglot.ast.For;
import polyglot.ast.If;
import polyglot.ast.Import;
import polyglot.ast.Local;
import polyglot.ast.LocalAssign;
import polyglot.ast.LocalDecl;
import polyglot.ast.MethodDecl;
import polyglot.ast.New;
import polyglot.ast.QualifierNode;
import polyglot.ast.Receiver;
import polyglot.ast.Return;
import polyglot.ast.SourceFile;
import polyglot.ast.Special;
import polyglot.ast.Stmt;
import polyglot.ast.Switch;
import polyglot.ast.Try;
import polyglot.ast.TypeNode;
import polyglot.ast.Unary;
import polyglot.ast.While;
import polyglot.ast.Import.Kind;
import polyglot.types.Flags;
import polyglot.util.Position;
import soot.javaToJimple.jj.ast.JjCast_c;
import soot.javaToJimple.jj.ast.JjNodeFactory_c;

/**
 * NodeFactory for ppa extension.
 */
public class PPANodeFactory_c extends JjNodeFactory_c implements PPANodeFactory {

	@Override
	public ArrayAccess ArrayAccess(Position pos, Expr base, Expr index) {
		ArrayAccess n = new PPAArrayAccess_c(pos, base, index);
		n = (ArrayAccess) n.ext(extFactory().extArrayAccess());
		n = (ArrayAccess) n.del(delFactory().delArrayAccess());
		return n;
	}
	
	@Override
	public ArrayAccessAssign ArrayAccessAssign(Position pos, ArrayAccess left, Assign.Operator op,
			Expr right) {
		ArrayAccessAssign n = new PPAArrayAccessAssign_c(pos, left, op, right);
		n = (ArrayAccessAssign) n.ext(extFactory().extArrayAccessAssign());
		n = (ArrayAccessAssign) n.del(delFactory().delArrayAccessAssign());
		return n;
	}

	@Override
	public ArrayInit ArrayInit(Position pos, List elements) {
		ArrayInit n = new PPAArrayInit_c(pos, elements);
		n = (ArrayInit) n.ext(extFactory().extArrayInit());
		n = (ArrayInit) n.del(delFactory().delArrayInit());
		return n;
	}

	@Override
	public Assign Assign(Position pos, Expr left, Assign.Operator op, Expr right) {
		// Assign n;
		if (left instanceof Local) {
			return LocalAssign(pos, (Local) left, op, right);
		} else if (left instanceof Field) {
			return FieldAssign(pos, (Field) left, op, right);
		} else if (left instanceof ArrayAccess) {
			return ArrayAccessAssign(pos, (ArrayAccess) left, op, right);
		}
		return AmbAssign(pos, left, op, right);
	}

	@Override
	public Binary Binary(Position pos, Expr left, Binary.Operator op, Expr right) {
		Binary n = new PPABinary_c(pos, left, op, right);
		n = (Binary) n.ext(extFactory().extBinary());
		n = (Binary) n.del(delFactory().delBinary());
		return n;
	}

	@Override
	public Call Call(Position pos, Receiver target, String name, List args) {
		Call n = new PPACall_c(pos, target, name, args);
		n = (Call) n.ext(extFactory().extCall());
		n = (Call) n.del(delFactory().delCall());
		return n;
	}

	@Override
	public Case Case(Position pos, Expr expr) {
		Case n = new PPACase_c(pos, expr);
		n = (Case) n.ext(extFactory().extCase());
		n = (Case) n.del(delFactory().delCase());
		return n;
	}

	@Override
	public Cast Cast(Position pos, TypeNode type, Expr expr) {
		Cast n = new JjCast_c(pos, type, expr);
		n = (Cast) n.ext(extFactory().extCast());
		n = (Cast) n.del(delFactory().delCast());
		return n;
	}

	@Override
	public ClassDecl ClassDecl(Position pos, Flags flags, String name, TypeNode superClass,
			List interfaces, ClassBody body) {
		ClassDecl n = new PPAClassDecl_c(pos, flags, name, superClass, interfaces, body);
		n = (ClassDecl) n.ext(extFactory().extClassDecl());
		n = (ClassDecl) n.del(delFactory().delClassDecl());
		return n;
	}

	@Override
	public Conditional Conditional(Position pos, Expr cond, Expr consequent, Expr alternative) {
		Conditional n = new PPAConditional_c(pos, cond, consequent, alternative);
		n = (Conditional) n.ext(extFactory().extConditional());
		n = (Conditional) n.del(delFactory().delConditional());
		return n;
	}

	@Override
	public ConstructorCall ConstructorCall(Position pos, polyglot.ast.ConstructorCall.Kind kind,
			Expr outer, List args) {
		ConstructorCall n = new PPAConstructorCall_c(pos, kind, outer, args);
		n = (ConstructorCall) n.ext(extFactory().extConstructorCall());
		n = (ConstructorCall) n.del(delFactory().delConstructorCall());
		return n;
	}

	@Override
	public ConstructorDecl ConstructorDecl(Position pos, Flags flags, String name, List formals,
			List throwTypes, Block body) {
		ConstructorDecl n = new PPAConstructorDecl_c(pos, flags, name, formals, throwTypes, body);
		n = (ConstructorDecl) n.ext(extFactory().extConstructorDecl());
		n = (ConstructorDecl) n.del(delFactory().delConstructorDecl());
		return n;
	}

	@Override
	public Disamb disamb() {
		return new PPADisamb_c();
	}

	@Override
	public AmbTypeNode AmbTypeNode(Position pos, QualifierNode qualifier, String name) {
		AmbTypeNode n = new PPAAmbTypeNode_c(pos, qualifier, name);
		n = (AmbTypeNode) n.ext(extFactory().extAmbTypeNode());
		n = (AmbTypeNode) n.del(delFactory().delAmbTypeNode());
		return n;
	}

	@Override
	public Do Do(Position pos, Stmt body, Expr cond) {
		Do n = new PPADo_c(pos, body, cond);
		n = (Do) n.ext(extFactory().extDo());
		n = (Do) n.del(delFactory().delDo());
		return n;
	}

	@Override
	public polyglot.ast.Field Field(Position pos, Receiver target, String name) {
		Field n = new PPAField_c(pos, target, name);
		n = (Field) n.ext(extFactory().extField());
		n = (Field) n.del(delFactory().delField());
		return n;
	}

	@Override
	public FieldAssign FieldAssign(Position pos, Field left, Assign.Operator op, Expr right) {
		FieldAssign n = new PPAFieldAssign_c(pos, left, op, right);
		n = (FieldAssign) n.ext(extFactory().extFieldAssign());
		n = (FieldAssign) n.del(delFactory().delFieldAssign());
		return n;
	}

	@Override
	public FieldDecl FieldDecl(Position pos, Flags flags, TypeNode type, String name, Expr init) {
		FieldDecl n = new PPAFieldDecl_c(pos, flags, type, name, init);
		n = (FieldDecl) n.ext(extFactory().extFieldDecl());
		n = (FieldDecl) n.del(delFactory().delFieldDecl());
		return n;
	}

	@Override
	public For For(Position pos, List inits, Expr cond, List iters, Stmt body) {
		For n = new PPAFor_c(pos, inits, cond, iters, body);
		n = (For) n.ext(extFactory().extFor());
		n = (For) n.del(delFactory().delFor());
		return n;
	}

	@Override
	public If If(Position pos, Expr cond, Stmt consequent, Stmt alternative) {
		If n = new PPAIf_c(pos, cond, consequent, alternative);
		n = (If) n.ext(extFactory().extIf());
		n = (If) n.del(delFactory().delIf());
		return n;
	}

	@Override
	public Import Import(Position pos, Kind kind, String name) {
		Import n = new PPAImport_c(pos, kind, name);
		n = (Import) n.ext(extFactory().extImport());
		n = (Import) n.del(delFactory().delImport());
		return n;
	}

	@Override
	public Local Local(Position pos, String name) {
		Local n = new PPALocal_c(pos, name);
		n = (Local) n.ext(extFactory().extLocal());
		n = (Local) n.del(delFactory().delLocal());
		return n;
	}

	@Override
	public LocalAssign LocalAssign(Position pos, Local left, Assign.Operator op, Expr right) {
		LocalAssign n = new PPALocalAssign_c(pos, left, op, right);
		n = (LocalAssign) n.ext(extFactory().extLocalAssign());
		n = (LocalAssign) n.del(delFactory().delLocalAssign());
		return n;
	}

	@Override
	public LocalDecl LocalDecl(Position pos, Flags flags, TypeNode type, String name, Expr init) {
		LocalDecl n = new PPALocalDecl_c(pos, flags, type, name, init);
		n = (LocalDecl) n.ext(extFactory().extLocalDecl());
		n = (LocalDecl) n.del(delFactory().delLocalDecl());
		return n;
	}

	@Override
	public MethodDecl MethodDecl(Position pos, Flags flags, TypeNode returnType, String name,
			List formals, List throwTypes, Block body) {
		MethodDecl n = new PPAMethodDecl_c(pos, flags, returnType, name, formals, throwTypes, body);
		n = (MethodDecl) n.ext(extFactory().extMethodDecl());
		n = (MethodDecl) n.del(delFactory().delMethodDecl());
		return n;
	}
	
	

	@Override
	public New New(Position pos, Expr outer, TypeNode objectType, List args, ClassBody body) {
		 New n = new PPANew_c(pos, outer, objectType, args, body);
	        n = (New)n.ext(extFactory().extNew());
	        n = (New)n.del(delFactory().delNew());
	        return n;
	}

	@Override
	public Return Return(Position pos, Expr expr) {
		Return n = new PPAReturn_c(pos, expr);
		n = (Return) n.ext(extFactory().extReturn());
		n = (Return) n.del(delFactory().delReturn());
		return n;
	}

	@Override
	public polyglot.ast.SourceFile SourceFile(Position pos, polyglot.ast.PackageNode packageName,
			List imports, List decls) {
		SourceFile n = new PPASourceFile_c(pos, packageName, imports, decls);
		n = (SourceFile) n.ext(extFactory().extSourceFile());
		n = (SourceFile) n.del(delFactory().delSourceFile());
		return n;
	}

	@Override
	public Special Special(Position pos, polyglot.ast.Special.Kind kind, TypeNode outer) {
		Special n = new PPASpecial_c(pos, kind, outer);
		n = (Special) n.ext(extFactory().extSpecial());
		n = (Special) n.del(delFactory().delSpecial());
		return n;
	}

	@Override
	public Switch Switch(Position pos, Expr expr, List elements) {
		Switch n = new PPASwitch(pos, expr, elements);
		n = (Switch) n.ext(extFactory().extSwitch());
		n = (Switch) n.del(delFactory().delSwitch());
		return n;
	}

	@Override
	public Try Try(Position pos, Block tryBlock, List catchBlocks, Block finallyBlock) {
		Try n = new PPATry_c(pos, tryBlock, catchBlocks, finallyBlock);
		n = (Try) n.ext(extFactory().extTry());
		n = (Try) n.del(delFactory().delTry());
		return n;
	}

	@Override
	public While While(Position pos, Expr cond, Stmt body) {
		While n = new PPAWhile_c(pos, cond, body);
		n = (While) n.ext(extFactory().extWhile());
		n = (While) n.del(delFactory().delWhile());
		return n;
	}

	@Override
	public Unary Unary(Position pos, Unary.Operator op, Expr expr) {
		Unary n = new PPAUnary_c(pos, op, expr);
		n = (Unary) n.ext(extFactory().extUnary());
		n = (Unary) n.del(delFactory().delUnary());
		return n;
	}

}
